//--------------------------------------------------------------------------------------
// Shader library header by TamaFutoshi
//--------------------------------------------------------------------------------------

static const uint SphereModeNone = 0;
static const uint SphereModeMul = 1;
static const uint SphereModeAdd = 2;
static const uint SphereModeSubTexture = 3;

struct VS_INPUT
{
    float3 Pos : POSITION0;
	float3 Normal: NORMAL0;
    float2 Tex : TEXCOORD0;
    uint4 boneRef: BONEREF0;
    float4 boneWeight: BONEWEIGHT0;
};

struct PS_INPUT
{
    float4 Pos : SV_POSITION;
	float4 Normal: NORMAL0;
    float2 Tex : TEXCOORD0;
    float4 WorldPos: WORLDPOSITION;
    float4 WorldNormal: WORLDNORMAL;
};

//order by change frequency
cbuffer ConstantBuffer : register(b0)  //Material constant buffer
{
	float4 baseColor;
	float3 specularColor;
	float specularStrength;
	float3 ambientColor;
	float edgeSize;
	float4 edgeColor;
	float3 emissionColor;
	float emissionStrength;
	float metallic;
	float roughness;
	float reflective;
	bool receiveShadow;
	uint sphereMode;
	bool textureBind0;
	bool textureBind1;
	bool textureBind2;
	bool textureBind3;
	bool textureBind4;
	bool textureBind5;
	bool textureBind6;
	bool textureBind7;
};

cbuffer ConstantBuffer : register(b1)  //Constant buffer
{
	float4x4 mModel;
	float4x4 mView;
	float4x4 mProjection;
    float4x4 mMVP;
    float4 cameraPos;
	float4 lightPos;
	float4 lightDir;
	float4 lightColor;
    float4x4 lightView;
    float4x4 lightProjection;
	float elapsedTime;
};

Texture2D g_texture0     : register(t0);  //PMX bind: Main texture
Texture2D g_texture1     : register(t1);  //PMX bind: Sphere texture
Texture2D g_texture2     : register(t2);  //PMX bind: Toon texture
Texture2D g_texture3     : register(t3);  //Reservered
Texture2D g_texture4     : register(t4);  //Reservered
Texture2D g_texture5     : register(t5);  //Reservered
Texture2D g_texture6     : register(t6);  //Reservered
Texture2D g_texture7     : register(t7);  //Reservered
Texture2D g_texture8     : register(t8);  //TFE bind: Shadow map texture
SamplerState g_samWarp   : register(s0);  //Address mode warp
SamplerState g_samClamp  : register(s1);  //Address mode clamp
SamplerState g_samBorder : register(s2);  //Address mode border 1.0f

bool IsBlack(float3 color)
{
	return !any(color);
}

//--------------------------------------------------------------------------------------
// Calculate unshadowed amount PCF2x2
//--------------------------------------------------------------------------------------
float4 CalculateUnshadowedAmount(float4 worldPos)
{
    //calculate position in light space
    float4 lightSpacePos = worldPos;
    lightSpacePos = mul(lightSpacePos, lightView);
    lightSpacePos = mul(lightSpacePos, lightProjection);

    lightSpacePos.xyz /= lightSpacePos.w;

    //Calculate texture coords
    float2 shadowTexCoord = 0.5f * lightSpacePos.xy + 0.5f;
    shadowTexCoord.y = 1.0f - shadowTexCoord.y;

    //Depth bias to avoid pixel self-shadowing
    float lightSpaceDepth = lightSpacePos.z - 0.00005f;
    
    //Find sub-pixel weights
    float shadowMapWidth = 1.0f;
    float shadowMapHeight = 1.0f;
    g_texture8.GetDimensions(shadowMapWidth, shadowMapHeight);
    float2 shadowMapDim = float2(shadowMapWidth, shadowMapHeight);
    float4 subPixelCoords = float4(1.0f, 1.0f, 1.0f, 1.0f);
    subPixelCoords.xy = frac(shadowMapDim * shadowTexCoord);
    subPixelCoords.zw = 1.0f - subPixelCoords.xy;
    float4 biLinearWeights = subPixelCoords.zxzx * subPixelCoords.wwyy;

    //2x2 percentage closer filtering
    float2 texelUnits = 1.0f / shadowMapDim;
    float4 shadowDepths;
    shadowDepths.x = g_texture8.Sample(g_samBorder, shadowTexCoord).r;
    shadowDepths.y = g_texture8.Sample(g_samBorder, shadowTexCoord + float2(texelUnits.x, 0.0f)).r;
    shadowDepths.z = g_texture8.Sample(g_samBorder, shadowTexCoord + float2(0.0f, texelUnits.y)).r;
    shadowDepths.w = g_texture8.Sample(g_samBorder, shadowTexCoord + texelUnits).r;

    //weights testing
    float4 shadowTests = (shadowDepths >= lightSpaceDepth) ? 1.0f : 0.0f;
    return dot(biLinearWeights, shadowTests);
}